@@ -2,8 +2,8 @@ from enum import Enum
 from PySide6.QtCore import Qt
 import pytest
 from dataclass2PySide6 import (type2Widget, BoolCheckBox, IntLineEdit,
-    FloatLineEdit, StrLineEdit, TupleGroupBox, EnumComboBox)
-from typing import Tuple
+    FloatLineEdit, StrLineEdit, TupleGroupBox, EnumComboBox, MISSING)
+from typing import Tuple, Union, Optional
 
 
 def test_type2Widget(qtbot):
@@ -27,104 +27,195 @@ def test_type2Widget(qtbot):
     assert isinstance(tuplegbox2.widgets()[1].widgets()[0], IntLineEdit)
 
 
+def test_type2Widget_Union(qtbot):
+    with pytest.raises(TypeError):
+        type2Widget(Union[int, float])
+
+    tristate_checkbox = type2Widget(Optional[bool])
+    assert isinstance(tristate_checkbox, BoolCheckBox)
+    assert tristate_checkbox.isTristate()
+
+    optint_checkbox = type2Widget(Optional[int])
+    assert isinstance(optint_checkbox, IntLineEdit)
+    assert optint_checkbox.defaultDataValue() is None
+
+    optfloat_checkbox = type2Widget(Optional[float])
+    assert isinstance(optfloat_checkbox, FloatLineEdit)
+    assert optfloat_checkbox.defaultDataValue() is None
+
+
 def test_BoolCheckBox(qtbot):
     widget = BoolCheckBox()
 
     # test dataValueChanged signal
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: val):
-        widget.setDataValue(True)
+                          check_params_cb=lambda val: val is True):
+        widget.setCheckState(Qt.Checked)
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: not val):
-        widget.setDataValue(False)
+                          check_params_cb=lambda val: val is False):
+        widget.setCheckState(Qt.Unchecked)
+
+    # test tristate
+    widget.setTristate(True)
+    with qtbot.waitSignal(widget.dataValueChanged,
+                          raising=True,
+                          check_params_cb=lambda val: val is True):
+        widget.setCheckState(Qt.Checked)
+    with qtbot.waitSignal(widget.dataValueChanged,
+                          raising=True,
+                          check_params_cb=lambda val: val is False):
+        widget.setCheckState(Qt.Unchecked)
+    with qtbot.waitSignal(widget.dataValueChanged,
+                          raising=True,
+                          check_params_cb=lambda val: val is None):
+        widget.setCheckState(Qt.PartiallyChecked)
 
 
 def test_IntLineEdit(qtbot):
     widget = IntLineEdit()
 
+    assert not widget.hasDefaultDataValue()
+    with pytest.raises(TypeError):
+        widget.dataValue()
+
     # test default data value
-    assert widget.dataValue() == 0
     widget.setDefaultDataValue(1)
+    assert widget.hasDefaultDataValue()
     assert widget.dataValue() == 1
-    widget.setDefaultDataValue(0)
+    widget.setDefaultDataValue(10)
+    assert widget.hasDefaultDataValue()
+    assert widget.dataValue() == 10
+    widget.setDefaultDataValue(None)
+    assert widget.hasDefaultDataValue()
+    assert widget.dataValue() == None
+    widget.setDefaultDataValue(MISSING)
+    assert not widget.hasDefaultDataValue()
+    with pytest.raises(TypeError):
+        widget.dataValue()
 
     # test dataValueChanged signal
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
                           check_params_cb=lambda val: val == 1):
-        widget.setDataValue("1")
+        qtbot.keyPress(widget, '1')
+        qtbot.keyPress(widget, Qt.Key_Return)
     assert widget.dataValue() == 1
+
+    widget.clear()
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: val == 0):
-        widget.setDataValue("")
-    assert widget.dataValue() == 0
+                          check_params_cb=lambda val: val == -1):
+        qtbot.keyPress(widget, '-')
+        qtbot.keyPress(widget, '1')
+        qtbot.keyPress(widget, Qt.Key_Return)
+    assert widget.dataValue() == -1
 
-    # test dataValueEdited signal
+    # dataValueChanged signal with empty text
     widget.clear()
-    with qtbot.waitSignal(widget.dataValueEdited,
+    with qtbot.assertNotEmitted(widget.dataValueChanged):
+        qtbot.keyPress(widget, Qt.Key_Return)
+    with pytest.raises(TypeError):
+        widget.dataValue()
+
+    widget.setDefaultDataValue(None)
+    with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: val == 1):
-        qtbot.mouseClick(widget, Qt.LeftButton)
-        qtbot.keyPress(widget, "1")
-    assert widget.dataValue() == 1
+                          check_params_cb=lambda val: val is None):
+        qtbot.keyPress(widget, Qt.Key_Return)
+
+    widget.setDefaultDataValue(10)
+    with qtbot.waitSignal(widget.dataValueChanged,
+                          raising=True,
+                          check_params_cb=lambda val: val == 10):
+        qtbot.keyPress(widget, Qt.Key_Return)
 
     # test validator
     widget.clear()
-    with qtbot.waitSignal(widget.dataValueEdited,
+    with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
                           check_params_cb=lambda val: val == 11):
-        qtbot.mouseClick(widget, Qt.LeftButton)
-        qtbot.keyPress(widget, "1")
-        qtbot.keyPress(widget, ".") # this key is ignored
-        qtbot.keyPress(widget, "1")
+        qtbot.keyPress(widget, '1')
+        qtbot.keyPress(widget, '.') # this key is ignored
+        qtbot.keyPress(widget, '1')
+        qtbot.keyPress(widget, Qt.Key_Return)
     assert widget.dataValue() == 11
 
 
 def test_FloatLineEdit(qtbot):
     widget = FloatLineEdit()
 
+    assert not widget.hasDefaultDataValue()
+    with pytest.raises(TypeError):
+        widget.dataValue()
+
     # test default data value
-    assert widget.dataValue() == float(0)
-    widget.setDefaultDataValue(float(1))
+    widget.setDefaultDataValue(1)
+    assert widget.hasDefaultDataValue()
     assert widget.dataValue() == float(1)
-    widget.setDefaultDataValue(float(0))
+    widget.setDefaultDataValue(10)
+    assert widget.hasDefaultDataValue()
+    assert widget.dataValue() == float(10)
+    widget.setDefaultDataValue(None)
+    assert widget.hasDefaultDataValue()
+    assert widget.dataValue() == None
+    widget.setDefaultDataValue(MISSING)
+    assert not widget.hasDefaultDataValue()
+    with pytest.raises(TypeError):
+        widget.dataValue()
 
     # test dataValueChanged signal
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
                           check_params_cb=lambda val: val == 1.2):
-        widget.setDataValue("1.2")
+        qtbot.keyPress(widget, '1')
+        qtbot.keyPress(widget, '.')
+        qtbot.keyPress(widget, '2')
+        qtbot.keyPress(widget, Qt.Key_Return)
     assert widget.dataValue() == 1.2
+
+    widget.clear()
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: val == float(0)):
-        widget.setDataValue("")
-    assert widget.dataValue() == float(0)
+                          check_params_cb=lambda val: val == -1.2):
+        qtbot.keyPress(widget, '-')
+        qtbot.keyPress(widget, '1')
+        qtbot.keyPress(widget, '.')
+        qtbot.keyPress(widget, '2')
+        qtbot.keyPress(widget, Qt.Key_Return)
+    assert widget.dataValue() == -1.2
 
-    # test dataValueEdited signal
+    # dataValueChanged signal with empty text
     widget.clear()
-    with qtbot.waitSignal(widget.dataValueEdited,
+    with qtbot.assertNotEmitted(widget.dataValueChanged):
+        qtbot.keyPress(widget, Qt.Key_Return)
+    with pytest.raises(TypeError):
+        widget.dataValue()
+
+    widget.setDefaultDataValue(None)
+    with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: val == 1.2):
-        qtbot.mouseClick(widget, Qt.LeftButton)
-        qtbot.keyPress(widget, "1")
-        qtbot.keyPress(widget, ".")
-        qtbot.keyPress(widget, "2")
-    assert widget.dataValue() == 1.2
+                          check_params_cb=lambda val: val is None):
+        qtbot.keyPress(widget, Qt.Key_Return)
+
+    widget.setDefaultDataValue(float(10))
+    with qtbot.waitSignal(widget.dataValueChanged,
+                          raising=True,
+                          check_params_cb=lambda val: val == float(10)):
+        qtbot.keyPress(widget, Qt.Key_Return)
 
     # test validator
     widget.clear()
-    with qtbot.waitSignal(widget.dataValueEdited,
+    with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
                           check_params_cb=lambda val: val == 1.2):
         qtbot.mouseClick(widget, Qt.LeftButton)
-        qtbot.keyPress(widget, "1")
-        qtbot.keyPress(widget, ".")
-        qtbot.keyPress(widget, ".") # this key is ignored
-        qtbot.keyPress(widget, "2")
+        qtbot.keyPress(widget, '1')
+        qtbot.keyPress(widget, '.')
+        qtbot.keyPress(widget, '.') # this key is ignored
+        qtbot.keyPress(widget, '2')
+        qtbot.keyPress(widget, Qt.Key_Return)
     assert widget.dataValue() == 1.2
 
 
@@ -134,31 +225,35 @@ def test_StrLineEdit(qtbot):
     # test dataValueChanged signal
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: val == "foo"):
-        widget.setDataValue("foo")
-    assert widget.dataValue() == "foo"
-
-    # test dataValueEdited signal
-    widget.clear()
-    with qtbot.waitSignal(widget.dataValueEdited,
+                          check_params_cb=lambda val: val == 'foo'):
+        qtbot.keyPress(widget, 'f')
+        qtbot.keyPress(widget, 'o')
+        qtbot.keyPress(widget, 'o')
+        qtbot.keyPress(widget, Qt.Key_Return)
+    assert widget.dataValue() == 'foo'
+    with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
-                          check_params_cb=lambda val: val == "bar"):
-        qtbot.mouseClick(widget, Qt.LeftButton)
-        qtbot.keyPress(widget, "b")
-        qtbot.keyPress(widget, "a")
-        qtbot.keyPress(widget, "r")
-    assert widget.dataValue() == "bar"
+                          check_params_cb=lambda val: val == ''):
+        widget.setText('')
+        qtbot.keyPress(widget, Qt.Key_Return)
+    assert widget.dataValue() == ''
 
 
 def test_TupleGroupBox(qtbot):
     widgets = [IntLineEdit(), FloatLineEdit()]
     widget = TupleGroupBox.fromWidgets(widgets)
 
+    with pytest.raises(TypeError):
+        widget.dataValue()
+
     # test dataValueChanged signal
     with qtbot.waitSignal(widget.dataValueChanged,
                           raising=True,
                           check_params_cb=lambda val: val == (42, 0.0)):
-        widget.widgets()[0].setDataValue(42)
+        widget.widgets()[0].setText('42')
+        qtbot.keyPress(widget.widgets()[0], Qt.Key_Return)
+        widget.widgets()[1].setText('0')
+        qtbot.keyPress(widget.widgets()[1], Qt.Key_Return)
     assert widget.dataValue() == (42, 0.0)
 
 
